@@ -2,95 +2,67 @@
 
 #
 #   Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
-#   All rights reserved. Copyright (c) 2016 VECTIONEER.
+#   All rights reserved. Copyright (c) 2016-2020 VECTIONEER.
 #
 
 from motorcortex.reply import Reply
 
 import os
+import queue
 import ctypes
-
 from concurrent.futures import ThreadPoolExecutor
-from nanomsg import wrapper as nn
-from nanomsg import REQ, SOL_SOCKET, RCVTIMEO, RCVMAXSIZE, AF_SP, create_writable_buffer
-from builtins import bytes
+from pynng import Req0, TLSConfig
 
 
 class Request(object):
-    """Request/Reply communication is used to send commands to a motorcortex server.
-
-        Args:
-            protobuf_types(MessageTypes): reference to a MessageTypes instance
-            parameter_tree(ParameterTree): reference to a ParameterTree instance
-    """
 
     def __init__(self, protobuf_types, parameter_tree):
+        self.__socket = None
+        self.__url = None
+        self.__connected_lock = None
         self.__protobuf_types = protobuf_types
         self.__parameter_tree = parameter_tree
-        self.__socket = None
-        self.__pool = ThreadPoolExecutor(1)
+        self.__pool = ThreadPoolExecutor()
 
-    def connect(self, url, timeout_ms=0):
-        """Opens a request connection.
+    def url(self):
+        return self.__url
 
-            Args:
-                url(str): motorcortex server URL
-                timeout_ms(int): request timeout in milliseconds
+    def connect(self, url, **kwargs):
+        conn_timeout_ms, recv_timeout_ms, certificate = self.parse(**kwargs)
+        self.__url = url
+        tls_config = None
+        if certificate:
+            tls_config = TLSConfig(TLSConfig.MODE_CLIENT, ca_files=certificate)
 
-            Returns:
-                bool: True - if connected, False otherwise
+        self.__socket = Req0(recv_timeout=recv_timeout_ms, tls_config=tls_config)
+        self.__connected_lock = queue.Queue()
 
-        """
+        def pre_connect_cb(pipe):
+            self.__connected_lock.put(True)
 
-        socket = nn.nn_socket(AF_SP, REQ)
+        def post_remove_cb(pipe):
+            self.__connected_lock.put(False)
 
-        # setting receive buffer size to inf
-        recv_size = bytes([0xff, 0xff, 0xff, 0xff])
-        nn.nn_setsockopt(socket, SOL_SOCKET, RCVMAXSIZE, recv_size)
+        self.__socket.add_pre_pipe_connect_cb(pre_connect_cb)
+        self.__socket.add_post_pipe_remove_cb(post_remove_cb)
+        self.__socket.dial(url)
 
-        if timeout_ms > 0:
-            timeout_buf = bytes(
-                [timeout_ms & 0xff, (timeout_ms >> 8) & 0xff, (timeout_ms >> 16) & 0xff, (timeout_ms >> 24) & 0xff])
-            nn.nn_setsockopt(socket, SOL_SOCKET, RCVTIMEO, timeout_buf)
-
-        if (nn.nn_connect(socket, url) == 1):
-            self.__socket = socket
-            return True
-
-        return False
+        return Reply(self.__pool.submit(self.waitForConnection, self.__connected_lock,
+                                        conn_timeout_ms / 1000.0))
 
     def close(self):
-        """Closes connection to the server"""
-
-        nn.nn_close(self.__socket)
-        self.__socket = None
-        print("Reply connection closed")
+        if self.__connected_lock:
+            self.__connected_lock.put(False)
+        self.__socket.close()
 
     def send(self, encoded_msg, do_not_decode_reply=False):
-        """Sends an encoded request to the server
-
-            Args:
-                encoded_msg(list(bytes)): binary array with encoded message
-                do_not_decode_reply(bool): if True reply is returned as a binary buffer
-
-            Returns:
-                Reply: A promise, which completes when reply from the server is received.
-                If request fails, catch callback is triggered.
-
-            Examples:
-                >>> msg = motorcortex.createType('motorcortex.LoadMsg')
-                >>> msg.file_name = 'control.xml'
-                >>> reply = req.send(req.encode(msg))
-                >>> reply_value = reply.get()
-        """
-
         if self.__socket is not None:
-            return Reply(self.__pool.submit(self.__send, encoded_msg, do_not_decode_reply))
-
+            return Reply(self.__pool.submit(self.__send, self.__socket, encoded_msg,
+                                            None if do_not_decode_reply else self.__protobuf_types))
         return None
 
     def login(self, login, password):
-        """Sends a login request to the server
+        """Send a login request to the server
 
             Args:
                 login(str): user login
@@ -101,7 +73,7 @@ class Request(object):
                 Returned message has a status code, which indicates a status of the login.
 
             Examples:
-                >>> login_reply = req.login('operator', 'god123');
+                >>> login_reply = req.login('operator', 'iddqd')
                 >>> login_msg = login_reply.get()
                 >>> if login_msg.status == motorcortex_msg.OK
                 >>>     print('User logged-in')
@@ -115,7 +87,7 @@ class Request(object):
         return self.send(self.__protobuf_types.encode(login_msg))
 
     def getParameterTreeHash(self):
-        """Requests a parameter tree hash from the server.
+        """Request a parameter tree hash from the server.
 
             Returns:
                 Reply(ParameterTreeMsg): A Promise, which resolves when parameter tree hash is received or fails
@@ -134,7 +106,7 @@ class Request(object):
         return self.send(self.__protobuf_types.encode(param_tree_hash_msg))
 
     def getParameterTree(self):
-        """Requests a parameter tree from the server.
+        """Request a parameter tree from the server.
 
             Returns:
                 Reply(ParameterTreeMsg): A Promise, which resolves when parameter tree is received or fails
@@ -147,22 +119,8 @@ class Request(object):
 
         """
 
-        hash = self.getParameterTreeHash().get()
-        tree = self.__loadParameterTreeFile(str(hash.hash))
-        if self.__loadParameterTreeFile(str(hash.hash)):
-            if hash.hash == self.__hash(tree):
-                print('Found parameter tree in the cache')
-                return Reply(self.__pool.submit(lambda tree: tree, tree))
-            else:
-                print('Found parameter tree but the cache does not match')
-        else:
-            print('Failed to find parameter tree in the cache')
-
-        # getting and instantiating data type from the loaded dict
-        param_tree_msg = self.__protobuf_types.createType('motorcortex.GetParameterTreeMsg')
-        handle = self.send(self.__protobuf_types.encode(param_tree_msg))
-        # encoding and sending data
-        return Reply(self.__pool.submit(lambda handle: self.__saveParameterTreeFile(handle.get()), handle))
+        return Reply(self.__pool.submit(self.__getParameterTree,
+                                        self.getParameterTreeHash(), self.__protobuf_types, self.__socket))
 
     def save(self, path, file_name):
         """Request a server to save a parameter tree to file.
@@ -178,13 +136,13 @@ class Request(object):
         """
 
         param_save_msg = self.__protobuf_types.createType('motorcortex.SaveMsg')
-        param_save_msg.path = path;
-        param_save_msg.file_name = file_name;
+        param_save_msg.path = path
+        param_save_msg.file_name = file_name
 
         return self.send(self.__protobuf_types.encode(param_save_msg))
 
     def setParameter(self, path, value, type_name=None):
-        """Sets new value to a parameter with given path
+        """Set new value to a parameter with given path
 
             Args:
                 path(str): parameter path in the tree
@@ -195,16 +153,18 @@ class Request(object):
                   Reply(StatusMsg): A Promise, which resolves when parameter value is updated or fails otherwise.
 
             Examples:
-                  >>> reply = req.setParameter("root/Control/activateSemiAuto", False);
-                  >>> reply.get();
-                  >>> reply = req.setParameter("root/Control/targetJointAngles", [0.2, 3.14, 0.4]);
+                  >>> reply = req.setParameter("root/Control/activateSemiAuto", False)
+                  >>> reply.get()
+                  >>> reply = req.setParameter("root/Control/targetJointAngles", [0.2, 3.14, 0.4])
                   >>> reply.get()
         """
 
-        return self.send(self.__protobuf_types.encode(self.__buildSetParameterMsg(path, value, type_name)))
+        return self.send(self.__protobuf_types.encode(self.__buildSetParameterMsg(path, value,
+                                                                                  type_name, self.__protobuf_types,
+                                                                                  self.__parameter_tree)))
 
     def setParameterList(self, param_list):
-        """Sets new values to a parameter list
+        """Set new values to a parameter list
 
             Args:
                  param_list([{'path'-`str`,'value'-`any`}]): a list of the parameters which values update
@@ -227,16 +187,18 @@ class Request(object):
             type_name = None
             if "type_name" in param:
                 type_name = param["type_name"]
-            set_param_list_msg.params.extend([self.__buildSetParameterMsg(param["path"], param["value"], type_name)])
+            set_param_list_msg.params.extend([self.__buildSetParameterMsg(param["path"], param["value"],
+                                                                          type_name, self.__protobuf_types,
+                                                                          self.__parameter_tree)])
 
         # encoding and sending data
         return self.send(self.__protobuf_types.encode(set_param_list_msg))
 
     def getParameter(self, path):
-        """Requests a parameter with description and value from the server.
+        """Request a parameter with description and value from the server.
 
             Args:
-                path_list(str): parameter path in the tree.
+                path(str): parameter path in the tree.
 
             Returns:
                  Reply(ParameterMsg): Returns a Promise, which resolves when parameter
@@ -247,7 +209,7 @@ class Request(object):
                 >>> param_full = param_reply.get() # Value and description
         """
 
-        return self.send(self.__protobuf_types.encode(self.__buildGetParameterMsg(path)))
+        return self.send(self.__protobuf_types.encode(self.__buildGetParameterMsg(path, self.__protobuf_types)))
 
     def getParameterList(self, path_list):
         """Get description and values of requested parameters.
@@ -269,13 +231,13 @@ class Request(object):
         get_param_list_msg = self.__protobuf_types.createType('motorcortex.GetParameterListMsg')
         # filling with sub messages
         for path in path_list:
-            get_param_list_msg.params.extend([self.__buildGetParameterMsg(path)])
+            get_param_list_msg.params.extend([self.__buildGetParameterMsg(path, self.__protobuf_types)])
 
         # encoding and sending data
         return self.send(self.__protobuf_types.encode(get_param_list_msg))
 
     def createGroup(self, path_list, group_alias, frq_divider=1):
-        """Creates a subscription group for a list of the parameters.
+        """Create a subscription group for a list of the parameters.
 
             This method is used inside Subscription class, use subscription class instead.
 
@@ -298,7 +260,7 @@ class Request(object):
         return self.send(self.__protobuf_types.encode(create_group_msg))
 
     def removeGroup(self, group_alias):
-        """Unsubscribes from the group.
+        """Unsubscribe from the group.
 
             This method is used inside Subscription class, use subscription class instead.
 
@@ -314,64 +276,121 @@ class Request(object):
         remove_group_msg = self.__protobuf_types.createType('motorcortex.RemoveGroupMsg')
         remove_group_msg.alias = group_alias
         # encoding and sending data
-        return self.send(self.__protobuf_types.encode(remove_group_msg));
+        return self.send(self.__protobuf_types.encode(remove_group_msg))
 
-    def __buildSetParameterMsg(self, path, value, type_name):
+    @staticmethod
+    def __buildSetParameterMsg(path, value, type_name, protobuf_types, parameter_tree):
         param_value = None
         if not type_name:
-            type_id = self.__parameter_tree.getDataType(path)
+            type_id = parameter_tree.getDataType(path)
             if type_id:
-                param_value = self.__protobuf_types.getTypeByHash(type_id)
+                param_value = protobuf_types.getTypeByHash(type_id)
         else:
-            param_value = self.__protobuf_types.createType(type_name)
+            param_value = protobuf_types.createType(type_name)
 
         if not param_value:
             print("Failed to find encoder for the path: %s type: %s" % (path, type_name))
 
         # creating type instance
-        set_param_msg = self.__protobuf_types.createType("motorcortex.SetParameterMsg")
+        set_param_msg = protobuf_types.createType("motorcortex.SetParameterMsg")
         set_param_msg.path = path
         # encoding parameter value
         set_param_msg.value = param_value.encode(value)
 
         return set_param_msg
 
-    def __buildGetParameterMsg(self, path):
+    @staticmethod
+    def __buildGetParameterMsg(path, protobuf_types):
         # getting and instantiating data type from the loaded dict
-        get_param_msg = self.__protobuf_types.createType('motorcortex.GetParameterMsg')
+        get_param_msg = protobuf_types.createType('motorcortex.GetParameterMsg')
         get_param_msg.path = path
 
         return get_param_msg
 
-    def __hash(self, tree):
-        hash = int()
+    @staticmethod
+    def parse(**kwargs):
+        conn_timeout_ms = 0
+        recv_timeout_ms = None
+        certificate = None
+        for key, value in kwargs.items():
+            if key == "recv_timeout_ms":
+                recv_timeout_ms = value
+            elif key == "conn_timeout_ms":
+                conn_timeout_ms = value
+            elif key == "timeout_ms":
+                conn_timeout_ms = value
+            elif key == 'certificate':
+                certificate = value
+            else:
+                print("Unknown parameter %s" % key)
+
+        return conn_timeout_ms, recv_timeout_ms, certificate
+
+    @staticmethod
+    def __send(req, encoded_msg, protobuf_types):
+        req.send(encoded_msg)
+        buffer = req.recv()
+        if buffer:
+            if protobuf_types:
+                return protobuf_types.decode(buffer)
+            else:
+                return protobuf_types
+
+        return None
+
+    @staticmethod
+    def waitForConnection(connected_lock, timeout_sec):
+        if timeout_sec <= 0:
+            timeout_sec = None
+        try:
+            return connected_lock.get(block=True, timeout=timeout_sec)
+        except queue.Empty:
+            return False
+
+    @staticmethod
+    def __hash(tree):
+        tree_hash = int()
+
+        def clip_u32(val):
+            return ctypes.c_uint32(val).value
+
         for param in tree.params:
-            clip_u32 = lambda val: ctypes.c_uint32(val).value
-            hash += clip_u32(param.data_size + param.data_type + param.id + param.number_of_elements + \
-                             param.param_type + param.unit + param.group_id + param.permissions)
+            tree_hash += clip_u32(param.data_size + param.data_type + param.id + param.number_of_elements +
+                                  param.param_type + param.unit + param.group_id + param.permissions)
             for ch in param.path:
-                hash = clip_u32(((hash << 5) - hash) + ord(ch))
-                hash |= 0
+                tree_hash = clip_u32(((tree_hash << 5) - tree_hash) + ord(ch))
+                tree_hash |= 0
 
             if hasattr(param, 'module_type'):
                 for ch in param.module_type:
-                    hash = (clip_u32(hash << 3) - hash) + ord(ch)
+                    tree_hash = (clip_u32(tree_hash << 3) - tree_hash) + ord(ch)
 
-        return hash
+        return tree_hash
 
-    def __loadParameterTreeFile(self, path):
-        if os.path.exists(path):
-            text_file = open(path, "rb")
-            data = text_file.read()
-            text_file.close()
-            param_tree_hash_msg = self.__protobuf_types.createType('motorcortex.ParameterTreeMsg')
-            param_tree_hash_msg.ParseFromString(data)
-            return param_tree_hash_msg
-        return None
+    @staticmethod
+    def __getParameterTree(hash_reply, protobuf_types, socket):
+        tree_hash = hash_reply.get()
+        tree = Request.__loadParameterTreeFile(str(tree_hash.hash), protobuf_types)
+        if tree:
+            if tree_hash.hash == Request.__hash(tree):
+                print('Found parameter tree in the cache')
+                return tree
+            else:
+                print('Found parameter tree but the cache does not match')
+        else:
+            print('Failed to find parameter tree in the cache')
 
-    def __saveParameterTreeFile(self, parameter_tree):
+        # getting and instantiating data type from the loaded dict
+        param_tree_msg = protobuf_types.createType('motorcortex.GetParameterTreeMsg')
+        handle = Request.__send(socket, protobuf_types.encode(param_tree_msg), protobuf_types)
+
+        # encoding and sending data
+        return Request.__saveParameterTreeFile(handle)
+
+    @staticmethod
+    def __saveParameterTreeFile(parameter_tree):
         print('Saved parameter tree to the cache')
-        cache = self.__hash(parameter_tree)
+        cache = Request.__hash(parameter_tree)
         path = str(cache)
         data = parameter_tree.SerializeToString()
         text_file = open(path, "wb")
@@ -379,18 +398,13 @@ class Request(object):
         text_file.close()
         return parameter_tree
 
-    def __cachedTree(self, tree):
-        return tree
-
-    def __send(self, encoded_msg, do_not_decode_reply):
-        socket = self.__socket
-        nn.nn_send(socket, encoded_msg, 0)
-        length, buffer = nn.nn_recv(socket, 0)
-        if length > 0:
-            if do_not_decode_reply:
-                return buffer
-            else:
-                return self.__protobuf_types.decode(bytes(buffer))
-
+    @staticmethod
+    def __loadParameterTreeFile(path, protobuf_types):
+        if os.path.exists(path):
+            text_file = open(path, "rb")
+            data = text_file.read()
+            text_file.close()
+            param_tree_hash_msg = protobuf_types.createType('motorcortex.ParameterTreeMsg')
+            param_tree_hash_msg.ParseFromString(data)
+            return param_tree_hash_msg
         return None
-
