@@ -5,15 +5,15 @@
 #   All rights reserved. Copyright (c) 2016 VECTIONEER.
 #
 
-
+from motorcortex.request import Request
+from motorcortex.request import Reply
 from motorcortex.subscription import Subscription
-from builtins import bytes
-import threading
-from nanomsg import wrapper as nn
-from nanomsg import SUB, SOL_SOCKET, RCVTIMEO, AF_SP, SUB_SUBSCRIBE, SUB_UNSUBSCRIBE, RCVBUF
+from pynng import Sub0, TLSConfig
+from concurrent.futures import ThreadPoolExecutor
+from queue import Queue
 
 
-class Subscribe(threading.Thread):
+class Subscribe:
     """Subscribe class is used to receive continuous parameter updates from motorcortex server.
     
         Subscribe class simplifies creating and removing subscription groups.
@@ -25,15 +25,16 @@ class Subscribe(threading.Thread):
     """
 
     def __init__(self, req, protobuf_types):
-        threading.Thread.__init__(self)
+        self.__socket = None
+        self.__connected_lock = None
+        self.__url = None
         self.__req = req
         self.__protobuf_types = protobuf_types
-        self.__socket = None
-        self.__stop = threading.Event()
         self.__subscriptions = dict()
+        self.__pool = ThreadPoolExecutor()
 
-    def connect(self, url):
-        """Opens a subscribe connection.
+    def connect(self, url, **kwargs):
+        """Open a subscribe connection.
 
             Args:
                 url(str): motorcortex server URL
@@ -42,49 +43,60 @@ class Subscribe(threading.Thread):
                 bool: True - if connected, False otherwise
         """
 
-        socket = nn.nn_socket(AF_SP, SUB)
-        if nn.nn_connect(socket, url) == 1:
-            self.__socket = socket
-            self.start()
-            return True
+        conn_timeout_ms, recv_timeout_ms, certificate = Request.parse(**kwargs)
+        if recv_timeout_ms == 0:
+            recv_timeout_ms = 500
 
-        return False
+        self.__url = url
+        tls_config = None
+        if certificate:
+            tls_config = TLSConfig(TLSConfig.MODE_CLIENT, ca_files=certificate)
+
+        self.__socket = Sub0(recv_timeout=recv_timeout_ms, tls_config=tls_config)
+
+        self.__connected_lock = Queue()
+
+        def pre_connect_cb(pipe):
+            self.__connected_lock.put(True)
+
+        def post_remove_cb(pipe):
+            self.__connected_lock.put(False)
+
+        self.__socket.add_pre_pipe_connect_cb(pre_connect_cb)
+        self.__socket.add_post_pipe_remove_cb(post_remove_cb)
+        self.__socket.dial(url)
+
+        self.__pool.submit(self.run, self.__socket)
+
+        return Reply(self.__pool.submit(Request.waitForConnection, self.__connected_lock,
+                                        conn_timeout_ms / 1000.0))
 
     def close(self):
-        """Closes connection to the server"""
-        self.__stop.set()
-        for id in list(self.__subscriptions):
-            self.__unsubscribe(id)
+        """Close connection to the server"""
+        if self.__connected_lock:
+            self.__connected_lock.put(False)
+        self.__socket.close()
+        self.__pool.shutdown(wait=True)
 
-        nn.nn_close(self.__socket)
-        self.join()
-
-    def run(self):
-        while not self.__stop.is_set():
-            length, buffer = nn.nn_recv(self.__socket, 0)
-            if (length > 0):
-                id_buf = bytes(buffer)[:4]
-                # id = id_buf[0] + (id_buf[1] << 8) + (id_buf[2] << 16) + (id_buf[3] << 24)
-                protocol_version = id_buf[3];
-                id = id_buf[0] + (id_buf[1] << 8) + (id_buf[2] << 16)
-                sub = self.__subscriptions.get(id)
+    def run(self, socket):
+        while self.__socket:
+            buffer = socket.recv()
+            if buffer:
+                sub_id_buf = buffer[:4]
+                protocol_version = sub_id_buf[3]
+                sub_id = sub_id_buf[0] + (sub_id_buf[1] << 8) + (sub_id_buf[2] << 16)
+                sub = self.__subscriptions.get(sub_id)
                 if sub:
+                    length = len(buffer)
                     if protocol_version == 1:
-                        sub._updateProtocol1(bytes(buffer)[4:], length - 4)
+                        sub._updateProtocol1(buffer[4:], length - 4)
                     elif protocol_version == 0:
-                        sub._updateProtocol0(bytes(buffer)[4:], length - 4)
+                        sub._updateProtocol0(buffer[4:], length - 4)
                     else:
                         print('Unknown protocol type: %d', protocol_version)
 
         print('Subscribe connection closed')
 
-    def __complete(self, msg, subscription):
-        if subscription._complete(msg):
-            # id_buf = bytes([msg.id & 0xff, (msg.id >> 8) & 0xff, (msg.id >> 16) & 0xff, (msg.id >> 24) & 0xff])
-            id_buf = bytes([msg.id & 0xff, (msg.id >> 8) & 0xff, (msg.id >> 16) & 0xff])
-            nn.nn_setsockopt(self.__socket, SUB, SUB_SUBSCRIBE, id_buf)
-            self.__subscriptions[msg.id] = subscription
-
     def subscribe(self, param_list, group_alias, frq_divider=1):
         """Create a subscription group for a list of the parameters.
 
@@ -101,13 +113,13 @@ class Subscribe(threading.Thread):
 
         subscription = Subscription(group_alias, self.__protobuf_types)
         reply = self.__req.createGroup(param_list, group_alias, frq_divider)
-        reply.then(lambda msg, subscription=subscription: self.__complete(msg, subscription)).catch(
+        reply.then(self.__complete, subscription, self.__socket).catch(
             subscription._failed)
 
-        return subscription;
+        return subscription
 
     def unsubscribe(self, subscription):
-        """Unsubscribes from the group.
+        """Unsubscribe from the group.
 
             Args:
                 subscription(Subscription): subscription handle
@@ -117,21 +129,20 @@ class Subscribe(threading.Thread):
                   operation is complete, fails otherwise.
 
         """
-
-        return self.__unsubscribe(subscription.id())
-
-    def __unsubscribe(self, id):
-
-        # id_buf = bytes([id & 0xff, (id >> 8) & 0xff, (id >> 16) & 0xff, (id >> 24) & 0xff])
-        id_buf = bytes([id & 0xff, (id >> 8) & 0xff, (id >> 16) & 0xff])
-        nn.nn_setsockopt(self.__socket, SUB, SUB_UNSUBSCRIBE, id_buf)
-
-        sub = self.__subscriptions[id]
+        sub_id = subscription.id()
+        sub_id_buf = bytes([sub_id & 0xff, (sub_id >> 8) & 0xff, (sub_id >> 16) & 0xff])
+        self.__socket.unsubscribe(sub_id_buf)
+        sub = self.__subscriptions[sub_id]
         # send remove group request to the server
         reply = self.__req.removeGroup(sub.alias())
         # stop sub update thread
         sub.done()
-
-        del self.__subscriptions[id]
+        del self.__subscriptions[sub_id]
 
         return reply
+
+    def __complete(self, msg, subscription, socket):
+        if subscription._complete(msg):
+            id_buf = bytes([msg.id & 0xff, (msg.id >> 8) & 0xff, (msg.id >> 16) & 0xff])
+            socket.subscribe(id_buf)
+            self.__subscriptions[msg.id] = subscription
